"""
Wrap the ``cp`` module allowing for managed SSH file transfers.

This module works by keeping a cachedir per SSH minion on the master.
Requested files are first cached there and afterwards replicated to
the minion using ``scp``. The returned paths will point to files in
the remote cachedir. You can convert these paths to the local ones
by calling ``cp.convert_cache_path``, which is a function unique
to the wrapper.

.. note::

    This wrapper currently has several limitations:

    * Replication will always be performed, even if the file exists
      in the minion cache dir in the correct state (no hash checks).
    * Even non-``salt://`` URIs will be fetched by the master node
      first in order for other wrappers to be able to employ this
      one for fetching remotes.
    * When replicating directories, they are currently not sent as
      a tar archive, but file per file, which is very inefficient.
    * You cannot transfer files from the minion to the master-side
      SSH minion cache, they will only be available on the remote.

.. note::
    For backwards-compatibility reasons, this wrapper currently does
    not behave the same as the execution module regarding ``saltenv``.
    The parameter defaults to ``base``, regardless of the current
    value of the minion setting.
"""

import logging
import os
import shlex
import urllib.parse
from pathlib import Path

import salt.client.ssh
import salt.fileclient
import salt.utils.files
import salt.utils.stringutils
import salt.utils.templates
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def _client():
    ckey = "_cp_shell"
    if ckey not in __context__:
        # Don't recreate the shell each time, the connection is closed
        # automatically after a command is sent.
        single = salt.client.ssh.Single(__opts__, "", **__salt__.kwargs)
        __context__[ckey] = single.shell
    return SSHCpClient(__opts__, __context__[ckey], __salt__.kwargs["id_"])


def _minion_visible_cache_path(client, path):
    """
    Normalize paths returned from ``SSHCpClient`` (master FS cache, staging
    tree, or already under the minion ``cachedir``) to ``__opts__["cachedir"]``
    layout on the SSH target.
    """
    if path in (None, False, True):
        return path
    retp = Path(path)
    _opts = globals().get("__opts__")
    if _opts is not None:
        minion_root = Path(_opts["cachedir"])
    else:
        # Unit tests import this module without loader-injected ``__opts__``;
        # ``SSHCpClient`` tests merge the minion cache root into ``client.opts``.
        minion_root = Path(client.opts["cachedir"])
    for base in (
        Path(client.get_cachedir(master=True)),
        Path(client.opts["cachedir"]),
        minion_root,
    ):
        try:
            return str(minion_root / retp.relative_to(base))
        except ValueError:
            continue
    return str(retp)


def get_file(path, dest, saltenv="base", makedirs=False, template=None, **kwargs):
    """
    Send a file from the fileserver to the specified location.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.get_file salt://path/to/file /minion/dest

    path
        The path on the fileserver, like ``salt://foo/bar.conf``. It is possible
        to specify the ``saltenv`` using the querystring syntax:
        ``salt://foo/bar.conf?saltenv=config``

    dest
        The absolute path to transfer the file to on the minion. If empty,
        the file will be cached in the minion's cache dir, under
        ``files/<saltenv>/<path>``.

    saltenv
        Salt fileserver environment from which to retrieve the file.
        Defaults to ``base``.

    makedirs
        Whether to create the parent directories for ``dest`` as needed.
        Defaults to false.

    template
        If ``path`` and ``dest`` parameters should be interpreted as templates,
        the name of the renderer to use.

        Template rendering can be enabled on both ``path`` and
        ``dest`` file paths like so:

        .. code-block:: bash

            salt-ssh '*' cp.get_file "salt://{{grains.os}}/vimrc" /etc/vimrc template=jinja

    Additional keyword arguments are passed through to the renderer, otherwise discarded.

    .. note::

        It may be necessary to quote the URL when using the querystring method,
        depending on the shell being used to run the command.

    .. note::

        gzip compression is not supported in the salt-ssh version of
        ``cp.get_file``.
    """
    gzip = kwargs.pop("gzip", None)
    if gzip is not None:
        log.warning("The gzip argument to cp.get_file in salt-ssh is unsupported")

    (path, dest) = _render_filenames(path, dest, saltenv, template, **kwargs)

    path, senv = salt.utils.url.split_env(path)
    if senv:
        saltenv = senv

    if not hash_file(path, saltenv):
        return ""
    else:
        with _client() as client:
            ret = client.get_file(path, dest, makedirs, saltenv)
            if not ret:
                return ret
            # Return the cache path on the minion, not the local one
            return _minion_visible_cache_path(client, client.target_map[ret])


def envs():
    """
    List available fileserver environments.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.envs
    """
    return __context__["fileclient"].envs()


def get_template(
    path, dest, template="jinja", saltenv="base", makedirs=False, **kwargs
):
    """
    Render a file as a template before writing it.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.get_template salt://path/to/template /minion/dest

    path
        The path on the fileserver, like ``salt://foo/bar.conf``. It is possible
        to specify the ``saltenv`` using the querystring syntax:
        ``salt://foo/bar.conf?saltenv=config``

    dest
        The absolute path to transfer the file to on the minion. If empty,
        the rendered template will be cached in the minion's cache dir,
        under ``extrn_files/<saltenv>/<path>``.

    template
        The renderer to use for rendering the template. Defaults to ``jinja``.

    saltenv
        The saltenv the template should be pulled from. Defaults to ``base``.

    makedirs
        Whether to create the parent directories for ``dest`` as needed.
        Defaults to false.

    Additional keyword arguments are passed verbatim to the renderer.
    """
    if "salt" not in kwargs:
        kwargs["salt"] = __salt__.value()
    if "pillar" not in kwargs:
        kwargs["pillar"] = __pillar__.value()
    if "grains" not in kwargs:
        kwargs["grains"] = __grains__.value()
    if "opts" not in kwargs:
        kwargs["opts"] = __opts__

    with _client() as client:
        ret = client.get_template(path, dest, template, makedirs, saltenv, **kwargs)
        if not ret:
            return ret
        # Return the cache path on the minion, not the local one
        return _minion_visible_cache_path(client, client.target_map[ret])


def get_dir(path, dest, saltenv="base", template=None, **kwargs):
    """
    Recursively transfer a directory from the fileserver to the minion.

    .. note::

        This can take a long time since each file is transferred separately
        currently.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.get_dir salt://path/to/dir/ /minion/dest

    path
        The path on the fileserver, like ``salt://foo/bar/``. It is possible
        to specify the ``saltenv`` using the querystring syntax:
        ``salt://foo/bar?saltenv=config``

    dest
        The absolute path to transfer the directory to on the minion. If empty,
        the directory will be cached in the minion's cache dir, under
        ``files/<saltenv>/<path>``. Note that parent directories will
        be created as required automatically.

    saltenv
        Salt fileserver environment from which to retrieve the directory.
        Defaults to ``base``.

    template
        If ``path`` and ``dest`` parameters should be interpreted as templates,
        the name of the renderer to use.

    .. note::

        gzip compression is not supported in the salt-ssh version of
        cp.get_dir. The argument is only accepted for interface compatibility.
    """
    # FIXME: transfer dirs using tar
    gzip = kwargs.pop("gzip", None)
    if gzip is not None:
        log.warning("The gzip argument to cp.get_dir in salt-ssh is unsupported")
    (path, dest) = _render_filenames(path, dest, saltenv, template, **kwargs)

    with _client() as client:
        ret = client.get_dir(path, dest, saltenv, gzip)
        if not ret:
            return ret
        # Return the cache path on the minion, not the local one
        return [client.target_map[x] for x in ret]


def get_url(path, dest="", saltenv="base", makedirs=False, source_hash=None):
    """
    Retrieve a single file from a URL.

    path
        A URL to download a file from. Supported URL schemes are: ``salt://``,
        ``http://``, ``https://``, ``ftp://``, ``s3://``, ``swift://`` and
        ``file://`` (local filesystem). If no scheme was specified, this is
        equivalent of using ``file://``.
        If a ``file://`` URL is given, the function just returns absolute path
        to that file on a local filesystem.
        The function returns ``False`` if Salt was unable to fetch a file from
        a ``salt://`` URL.

        .. note::

            The file:// scheme is currently only partially supported in salt-ssh.
            It behaves the same as the unwrapped ``cp.get_url`` if dest is not
            ``None``, but returning its contents will fail. Use ``get_file_str``
            as a workaround for text files.

    dest
        The destination to write the cached file to. If empty, will cache the file
        in the minion's cache dir under ``extrn_files/<saltenv>/<hostname>/<path>``.
        Defaults to empty (i.e. caching the file).

        .. note::

            To simply return the file contents instead, set destination to
            ``None``. This works with ``salt://``, ``http://`` and ``https://``
            URLs. The files fetched by ``http://`` and ``https://`` will not
            be cached.

    saltenv
        Salt fileserver environment from which to retrieve the file. Ignored if
        ``path`` is not a ``salt://`` URL. Defaults to ``base``.

    makedirs
        Whether to create the parent directories for ``dest`` as needed.
        Defaults to false.

    source_hash
        If ``path`` is an http(s) or ftp URL and the file exists in the
        minion's file cache, this option can be passed to keep the minion from
        re-downloading the file if the cached copy matches the specified hash.
    """
    with _client() as client:
        if isinstance(dest, str):
            result = client.get_url(
                path, dest, makedirs, saltenv, source_hash=source_hash
            )
        else:
            result = client.get_url(
                path, None, makedirs, saltenv, no_cache=True, source_hash=source_hash
            )
        if not result:
            log.error(
                "Unable to fetch file %s from saltenv %s.",
                salt.utils.url.redact_http_basic_auth(path),
                saltenv,
            )
            return result
        if isinstance(dest, str):
            # Return the cache path on the minion, not the local one
            result = _minion_visible_cache_path(client, client.target_map[result])
        return salt.utils.stringutils.to_unicode(result)


def get_file_str(path, saltenv="base"):
    """
    Download a file from a URL to the Minion cache directory and return the
    contents of that file.

    Returns ``False`` if Salt was unable to cache a file from a URL.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.get_file_str salt://my/file

    path
        The path on the fileserver, like ``salt://foo/bar/``. It is possible
        to specify the ``saltenv`` using the querystring syntax:
        ``salt://foo/bar?saltenv=config``

    saltenv
        Salt fileserver environment from which to retrieve the file.
    """
    fn_ = cache_file(path, saltenv)
    if isinstance(fn_, str):
        try:
            with salt.utils.files.fopen(fn_, "r") as fp_:
                return salt.utils.stringutils.to_unicode(fp_.read())
        except OSError:
            return False
    return fn_


def cache_file(path, saltenv="base", source_hash=None, verify_ssl=True, use_etag=False):
    """
    Cache a single file on the Minion.
    Returns the location of the new cached file on the Minion.
    If the path being cached is a ``salt://`` URI, and the path does not exist,
    then ``False`` will be returned.

    If the path refers to a fileserver path (``salt://`` URI) and this is a state run,
    the file will also be added to the package of files that's sent to the minion
    for executing the state run (this behaves like ``extra_filerefs``).

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.cache_file salt://path/to/file

    path
        The path on the fileserver, like ``salt://foo/bar/``. It is possible
        to specify the ``saltenv`` using the querystring syntax:
        ``salt://foo/bar?saltenv=config``

    saltenv
        Salt fileserver environment from which to retrieve the file. Ignored if
        ``path`` is not a ``salt://`` URL. Defaults to ``base``.

    source_hash
        If ``name`` is an http(s) or ftp URL and the file exists in the
        minion's file cache, this option can be passed to keep the minion from
        re-downloading the file if the cached copy matches the specified hash.

        .. versionadded:: 2018.3.0

    verify_ssl
        If ``False``, remote https file sources (``https://``) and source_hash
        will not attempt to validate the servers certificate. Default is True.

        .. versionadded:: 3002

    use_etag
        If ``True``, remote http/https file sources will attempt to use the
        ETag header to determine if the remote file needs to be downloaded.
        This provides a lightweight mechanism for promptly refreshing files
        changed on a web server without requiring a full hash comparison via
        the ``source_hash`` parameter.

        .. versionadded:: 3005

    .. note::
        You can instrumentalize this function in your ``sls`` files to workaround a
        limitation in how ``salt-ssh`` handles Jinja imports:

        Imports in templates that will be rendered on the minion (usually during
        ``file.managed`` calls) will fail since the corresponding file is not
        sent to the minion by default.

        By caching it explicitly in your states, you can ensure it will be included
        in the filerefs that will be sent to the minion.

        .. code-block:: jinja

            # /srv/salt/my/config.sls
            {%- do salt["cp.cache_file"]("salt://my/map.jinja") %}

            Serialize config:
              file.managed:
                - name: /etc/my/config.conf
                - source: salt://my/files/config.conf.j2
                - template: jinja

            # /srv/salt/my/config.conf.j2
            {%- from "my/map.jinja" import mapdata with context %}
            myconf = {{ mapdata["foo"] }}

            # /srv/salt/my/map.jinja
            {%- set mapdata = {"foo": "bar"} %}
    """
    path = salt.utils.data.decode(path)
    saltenv = salt.utils.data.decode(saltenv)

    url_data = urllib.parse.urlparse(path)
    if url_data.scheme in ("file", ""):
        return __salt__["cp.cache_file_ssh"](
            path,
            saltenv=saltenv,
            source_hash=source_hash,
            verify_ssl=verify_ssl,
            use_etag=use_etag,
        )

    contextkey = "{}_|-{}_|-{}".format("cp.cache_file", path, saltenv)
    filerefs_ckey = "_cp_extra_filerefs"
    url_data = urllib.parse.urlparse(path)
    path_is_remote = url_data.scheme in salt.utils.files.REMOTE_PROTOS

    def _check_return(result):
        if result and url_data.scheme == "salt":
            if filerefs_ckey not in __context__:
                __context__[filerefs_ckey] = []
            if path not in __context__[filerefs_ckey]:
                __context__[filerefs_ckey].append(path)
        return result

    with _client() as client:
        try:
            if path_is_remote and contextkey in __context__:
                # Prevent multiple caches in the same salt run. Affects remote URLs
                # since the master won't know their hash, so the fileclient
                # wouldn't be able to prevent multiple caches if we try to cache
                # the remote URL more than once.
                if client._path_exists(__context__[contextkey]):
                    return _check_return(__context__[contextkey])
                else:
                    # File is in __context__ but no longer exists in the minion
                    # cache, get rid of the context key and re-cache below.
                    # Accounts for corner case where file is removed from minion
                    # cache between cp.cache_file calls in the same salt-run.
                    __context__.pop(contextkey)
        except AttributeError:
            pass

        # saltenv split from path is performed in client
        result = client.cache_file(
            path,
            saltenv,
            source_hash=source_hash,
            verify_ssl=verify_ssl,
            use_etag=use_etag,
        )
        if not result and not use_etag:
            log.error("Unable to cache file '%s' from saltenv '%s'.", path, saltenv)
        if result:
            # Return the cache path on the minion, not the local one
            result = client.target_map[result]
        if path_is_remote:
            # Cache was successful, store the result in __context__ to prevent
            # multiple caches (see above).
            __context__[contextkey] = result
        return _check_return(result)


def cache_files(paths, saltenv="base"):
    """
    Used to gather many files from the Master, the gathered files will be
    saved in the minion cachedir reflective to the paths retrieved from the
    Master.

    .. note::
        This can take a long time since each file is transferred separately.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.cache_files salt://pathto/file1,salt://pathto/file1

    There are two ways of defining the fileserver environment (a.k.a.
    ``saltenv``) from which to cache the files. One is to use the ``saltenv``
    parameter, and the other is to use a querystring syntax in the ``salt://``
    URL. The below two examples are equivalent:

    .. code-block:: bash

        salt '*' cp.cache_files salt://foo/bar.conf,salt://foo/baz.conf saltenv=config
        salt '*' cp.cache_files salt://foo/bar.conf?saltenv=config,salt://foo/baz.conf?saltenv=config

    The querystring method is less useful when all files are being cached from
    the same environment, but is a good way of caching files from multiple
    different environments in the same command. For example, the below command
    will cache the first file from the ``config1`` environment, and the second
    one from the ``config2`` environment.

    .. code-block:: bash

        salt '*' cp.cache_files salt://foo/bar.conf?saltenv=config1,salt://foo/bar.conf?saltenv=config2

    .. note::
        It may be necessary to quote the URL when using the querystring method,
        depending on the shell being used to run the command.
    """
    # FIXME: transfer using tar
    ret = []
    if isinstance(paths, str):
        paths = paths.split(",")
    for path in paths:
        ret.append(cache_file(path, saltenv))
    return ret


def cache_dir(
    path, saltenv="base", include_empty=False, include_pat=None, exclude_pat=None
):
    """
    Download and cache everything under a directory from the master.

    .. note::
        This can take a long time since each file is transferred separately.

    CLI Example:

    .. code-block:: bash

        salt '*' cp.cache_dir salt://path/to/dir
        salt '*' cp.cache_dir salt://path/to/dir include_pat='E@*.py$'

    path
        The path on the fileserver, like ``salt://foo/bar/``. It is possible
        to specify the ``saltenv`` using the querystring syntax:
        ``salt://foo/bar?saltenv=config``

    saltenv
        Salt fileserver environment from which to retrieve the directory.
        Defaults to ``base``.

    include_empty
        Whether to cache empty directories as well. Defaults to false.

    include_pat : None
        Glob or regex to narrow down the files cached from the given path. If
        matching with a regex, the regex must be prefixed with ``E@``,
        otherwise the expression will be interpreted as a glob.

        .. versionadded:: 2014.7.0

    exclude_pat : None
        Glob or regex to exclude certain files from being cached from the given
        path. If matching with a regex, the regex must be prefixed with ``E@``,
        otherwise the expression will be interpreted as a glob.

        .. note::

            If used with ``include_pat``, files matching this pattern will be
            excluded from the subset of files defined by ``include_pat``.

        .. versionadded:: 2014.7.0
    """
    # FIXME: transfer using tar
    with _client() as client:
        ret = client.cache_dir(path, saltenv, include_empty, include_pat, exclude_pat)
        if not ret:
            return ret
        return [client.target_map[x] for x in ret]


def cache_master(saltenv="base"):
    """
    Retrieve all of the files on the master and cache them locally.

    .. note::
        This can take a long time since each file is transferred separately.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.cache_master
    """
    # FIXME: transfer using tar
    with _client() as client:
        ret = client.cache_master(saltenv)
        if not ret:
            return ret
        parsed = []
        for file in ret:
            try:
                parsed.append(client.target_map[file])
            except KeyError:
                # Usually because file is False. We can't easily know which one though
                log.error("Failed transferring a file")
        return parsed


def list_states(saltenv="base"):
    """
    List all of the available state files in an environment.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.list_states

    saltenv
        Salt fileserver environment from which to list states.
        Defaults to ``base``.
    """
    return __context__["fileclient"].list_states(saltenv)


def list_master(saltenv="base", prefix=""):
    """
    List all of the files stored on the master.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.list_master

    saltenv
        Salt fileserver environment from which to list files.
        Defaults to ``base``.

    prefix
        Only list files under this prefix. Defaults to empty.
    """
    return __context__["fileclient"].file_list(saltenv, prefix)


def list_master_dirs(saltenv="base", prefix=""):
    """
    List all of the directories stored on the master.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.list_master_dirs

    saltenv
        Salt fileserver environment from which to list directories.
        Defaults to ``base``.

    prefix
        Only list directories under this prefix. Defaults to empty.
    """
    return __context__["fileclient"].dir_list(saltenv, prefix)


def list_master_symlinks(saltenv="base", prefix=""):
    """
    List all of the symlinks stored on the master.
    Will return a mapping of symlink names to absolute paths.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.list_master_symlinks

    saltenv
        Salt fileserver environment from which to list symlinks.
        Defaults to ``base``.

    prefix
        Only list symlinks under this prefix. Defaults to empty.
    """
    return __context__["fileclient"].symlink_list(saltenv, prefix)


def is_cached(path, saltenv="base"):
    """
    Returns the full path to a file if it is cached locally on the minion
    as well as the SSH master-minion, otherwise returns a blank string.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.is_cached salt://path/to/file

    path
        The path to check.

    saltenv
        Salt fileserver environment the file was cached from.
        Defaults to ``base``.
    """
    # saltenv split is done in client
    with _client() as client:
        ret = client.is_cached(path, saltenv)
        if not ret:
            return ret
        return _minion_visible_cache_path(client, ret)


def hash_file(path, saltenv="base"):
    """
    Return the hash of a file. Supports ``salt://`` URIs and local files.
    Local files should be specified with their absolute paths, without the
    ``file://`` scheme.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.hash_file salt://path/to/file
        salt-ssh '*' cp.hash_file /path/to/file

    path
        The path to return the hash for.

    saltenv
        Salt fileserver environment from which the file should be hashed.
        Defaults to ``base``.
    """
    path, senv = salt.utils.url.split_env(path)
    if senv:
        saltenv = senv
    url_data = urllib.parse.urlparse(path)
    if url_data.scheme in ("file", ""):
        return __salt__["cp.hash_file_ssh"](path, saltenv)
    with _client() as client:
        return client.hash_file(path, saltenv)


def convert_cache_path(path, cachedir=None, master=True):
    """
    It converts a path received by caching a file to the minion cache to the
    corresponding one in the local master cache (or the other way around).

    .. note::

        This function is exclusive to the SSH wrapper module and is mostly
        intended for other wrapper modules to use, not on the CLI.

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' cp.convert_cache_path /var/tmp/.root_abc123_salt/running_data/var/cache/salt/minion/files/base/foo.txt

    path
        The path to convert. It has to be in one of the (remote or SSH master-minion)
        cachedirs to be converted, otherwise will be returned verbatim.

    cachedir
        An optional cachedir override that was used when caching the file.

    master
        Whether to convert the path to the master-side path. Defaults
        to true (since this module returns the minion paths otherwise).
    """
    with _client() as client:
        return str(client.convert_path(path, cachedir, master))


def _gather_pillar(pillarenv, pillar_override):
    """
    The opts used during pillar rendering should contain the master
    opts in the root namespace. self.opts is the modified minion opts,
    containing the original master opts in __master_opts__.
    """
    popts = {}
    # Pillar compilation needs the master opts primarily,
    # same as during regular operation.
    popts.update(__opts__)
    popts.update(__opts__.get("__master_opts__", {}))
    pillar = salt.pillar.get_pillar(
        popts,
        __grains__.value(),
        __salt__.kwargs["id_"],
        __opts__["saltenv"] or "base",
        pillar_override=pillar_override,
        pillarenv=pillarenv,
    )
    return pillar.compile_pillar()


def _render_filenames(path, dest, saltenv, template, **kw):
    """
    Process markup in the :param:`path` and :param:`dest` variables (NOT the
    files under the paths they ultimately point to) according to the markup
    format provided by :param:`template`.
    """
    if not template:
        return (path, dest)

    # render the path as a template using path_template_engine as the engine
    if template not in salt.utils.templates.TEMPLATE_REGISTRY:
        raise CommandExecutionError(
            f"Attempted to render file paths with unavailable engine {template}"
        )

    kwargs = {}
    kwargs["salt"] = __salt__.value()
    if "pillarenv" in kw or "pillar" in kw:
        pillarenv = kw.get("pillarenv", __opts__.get("pillarenv"))
        kwargs["pillar"] = _gather_pillar(pillarenv, kw.get("pillar"))
    else:
        kwargs["pillar"] = __pillar__.value()
    kwargs["grains"] = __grains__.value()
    kwargs["opts"] = __opts__
    kwargs["saltenv"] = saltenv

    def _render(contents):
        """
        Render :param:`contents` into a literal pathname by writing it to a
        temp file, rendering that file, and returning the result.
        """
        # write out path to temp file
        tmp_path_fn = salt.utils.files.mkstemp()
        with salt.utils.files.fopen(tmp_path_fn, "w+") as fp_:
            fp_.write(salt.utils.stringutils.to_str(contents))
        data = salt.utils.templates.TEMPLATE_REGISTRY[template](
            tmp_path_fn, to_str=True, **kwargs
        )
        salt.utils.files.safe_rm(tmp_path_fn)
        if not data["result"]:
            # Failed to render the template
            raise CommandExecutionError(
                "Failed to render file path with error: {}".format(data["data"])
            )
        else:
            return data["data"]

    path = _render(path)
    dest = _render(dest)
    return (path, dest)


class SSHCpClient(salt.fileclient.FSClient):
    """
    A FileClient that replicates between SSH master-minion and remote minion caches
    """

    def __init__(self, opts, shell, tgt):  # pylint: disable=W0231
        salt.fileclient.FSClient.__init__(self, opts)  # pylint: disable=W0233
        self.shell = shell
        self.tgt = tgt
        # Internally, we need to return master paths, but in the wrapper functions,
        # we usually want to return the effective path on the minion.
        # This client is used for a single execution, thus we can easily save
        # all affected file paths for a lookup later.
        self.target_map = {}

    def _local_path_exists(self, path):
        file = self.convert_path(path, master=True)
        return file.exists()

    def _remote_path_exists(self, path):
        # ensure it's the minion path
        path = self.convert_path(path)
        _, _, retcode = self.shell.exec_cmd("test -e " + shlex.quote(str(path)))
        return not retcode

    def _path_exists(self, path):
        return self._local_path_exists(path) and self._remote_path_exists(path)

    def cache_local_file(self, path, **kwargs):
        raise CommandExecutionError("Cannot cache local files via salt-ssh")

    def is_cached(self, path, saltenv="base", cachedir=None):
        """
        Returns the full path to a file if it is cached both locally on the
        SSH master-minion and the minion, otherwise returns a blank string
        """
        if path.startswith("salt://"):
            path, senv = salt.utils.url.parse(path)
            if senv:
                saltenv = senv

        escaped = True if salt.utils.url.is_escaped(path) else False

        # also strip escape character '|'
        localsfilesdest = os.path.join(
            self.opts["cachedir"], "localfiles", path.lstrip("|/")
        )
        filesdest = os.path.join(
            self.opts["cachedir"], "files", saltenv, path.lstrip("|/")
        )
        extrndest = self._extrn_path(path, saltenv, cachedir=cachedir)

        extrn_ok = self._path_exists(extrndest)
        files_ok = self._path_exists(filesdest)
        # If both a plain fileserver cache and a rendered-template cache exist,
        # prefer the template output path. Otherwise preserve the historical
        # order (files, then localsfiles, then extrn) so ``cp.cache_file`` hits
        # resolve to ``files/`` as before.
        if extrn_ok and files_ok:
            return extrndest
        if files_ok:
            return salt.utils.url.escape(filesdest) if escaped else filesdest
        # While we do not cache minion-local files back on the master,
        # we can inspect the minion cache dir remotely
        if self._remote_path_exists(localsfilesdest):
            return (
                salt.utils.url.escape(localsfilesdest) if escaped else localsfilesdest
            )
        if extrn_ok:
            return extrndest

        return ""

    def _master_staging_root(self):
        """
        Directory root for ``salt-ssh/<id>/`` file staging on the SSH target.

        When ``thin_dir`` is set, staging must live *next to* the thin tree, not
        under it: ``ssh_py_shim`` removes ``OPTIONS.saltdir`` (the thin dir) when
        ``ssh_wipe`` is enabled, which would otherwise delete fetched files before
        integration tests (or callers) can inspect them on ``localhost``.
        """
        thin_dir = self.opts.get("thin_dir")
        if not thin_dir:
            return self.opts["cachedir"]
        thin_dir = os.path.abspath(str(thin_dir).rstrip(os.sep))
        parent = os.path.dirname(thin_dir)
        return os.path.join(parent, os.path.basename(thin_dir) + ".salt-ssh-file-cache")

    def get_cachedir(
        self, cachedir=None, master=True
    ):  # pylint: disable=arguments-differ
        prefix = ["salt-ssh", self.tgt] if master else []
        root = self._master_staging_root() if master else self.opts["cachedir"]
        if cachedir is None:
            return os.path.join(root, *prefix)
        if not os.path.isabs(cachedir):
            return os.path.join(root, *prefix, cachedir)
        if master:
            # The root cachedir on the master-side should not be overridden
            return os.path.join(
                root,
                *prefix,
                "absolute_root",
                str(Path(*cachedir.split(os.sep)[1:])),
            )
        return cachedir

    def convert_path(self, path, cachedir=None, master=False):
        """
        Convert a cache path from master/minion to the other.

        Both use the same cachedir in salt-ssh, but our fileclient
        here caches to a subdir on the master. Remove/add it from/to
        the path.
        """
        path = Path(path)
        master_cachedir = Path(self.get_cachedir(cachedir, master=True))
        minion_cachedir = Path(self.get_cachedir(cachedir, master=False))
        if master:
            # ``Path.parents`` membership is a poor proxy for "under this dir":
            # e.g. ``/var/tmp/thin`` is an ancestor of ``thin/running_data/.../files``
            # even when the path is still the minion-side layout.
            if path.is_relative_to(master_cachedir):
                return path
            return master_cachedir / path.relative_to(minion_cachedir)
        if not path.is_relative_to(master_cachedir):
            return path
        return minion_cachedir / path.relative_to(master_cachedir)

    def _send_file(self, src, dest, makedirs, cachedir):
        def _error(stdout, stderr):
            log.error("Failed sending file: %s", stderr or stdout)
            if Path(src).is_relative_to(Path(self.get_cachedir(cachedir))):
                # remove the cached file if the transfer fails
                Path(src).unlink(missing_ok=True)
            return False

        for path in (src, dest):
            if not Path(path).is_absolute():
                raise ValueError(
                    f"Paths must be absolute, got '{path}' as {'src' if path == src else 'dest'}"
                )
        src, dest = str(src), str(dest)  # ensure we're using strings
        stdout, stderr, retcode = self.shell.send(src, dest, makedirs)
        if retcode and makedirs and "Not a directory" in stderr:
            # This means the file path contains a parent that is currently a file
            # Remove it, but only if it's in our cache dir
            minion_cachedir = Path(self.get_cachedir(cachedir, master=False))
            dest = cur = Path(dest)
            while minion_cachedir in cur.parents:
                if self._isfile(cur):
                    if not self._rmpath(cur):
                        return _error(stdout, stderr)
                    dest = str(dest)
                    break
                cur = cur.parent
            else:
                return _error(stdout, stderr)
            # The offending file was removed, retry
            stdout, stderr, retcode = self.shell.send(src, dest, makedirs)
        if retcode:
            return _error(stdout, stderr)
        self.target_map[src] = dest
        # we need to return the local source for internal functionality
        return src

    def _isdir(self, path):
        _, _, retcode = self.shell.exec_cmd("test -d " + shlex.quote(str(path)))
        return not retcode

    def _isfile(self, path):
        _, _, retcode = self.shell.exec_cmd("test -f " + shlex.quote(str(path)))
        return not retcode

    def _rmpath(self, path, cachedir=None):
        path = Path(path)
        if not path or not path.is_absolute() or str(path) == "/":
            raise ValueError(
                f"Not deleting unspecified, relative or root path: '{path}'"
            )
        minion_cachedir = Path(self.get_cachedir(cachedir, master=False))
        if not path.is_relative_to(minion_cachedir):
            raise ValueError(
                f"Not recursively deleting a path outside of the cachedir. Path: '{path}'"
            )
        stdout, stderr, retcode = self.shell.exec_cmd(
            "rm -rf " + shlex.quote(str(path))
        )
        if retcode:
            log.error("Failed deleting path '%s': %s", path, stderr or stdout)
        return not retcode

    def get_url(
        self,
        url,
        dest,
        makedirs=False,
        saltenv="base",
        no_cache=False,
        cachedir=None,
        source_hash=None,
        verify_ssl=True,
        use_etag=False,
    ):
        url_data = urllib.parse.urlparse(url)
        if url_data.scheme in ("file", ""):
            # This should be executed on the minion (unwrapped)
            log.error("The file:// scheme is not supported via the salt-ssh cp wrapper")
            return False
        # Ensure we don't send the file twice
        if url_data.scheme == "salt":
            result = self.get_file(url, dest, makedirs, saltenv, cachedir=cachedir)
            if result and dest is None:
                with salt.utils.files.fopen(result, "rb") as fp_:
                    data = fp_.read()
                return data
            return result
        cached = super().get_url(
            url,
            "",
            makedirs=True,
            saltenv=saltenv,
            no_cache=no_cache,
            cachedir=cachedir,
            source_hash=source_hash,
            verify_ssl=verify_ssl,
            use_etag=use_etag,
        )
        if not cached:
            return cached
        if not isinstance(dest, str) and no_cache:
            # Only if dest is None and no_cache is True, the contents
            # will be found in cached, otherwise the regular fileclient
            # behaves the same as with dest == ""
            return cached
        strict = False
        if not dest:
            # The file needs to be cached to the minion cache.
            # We're using the same cachedir on the ssh master and the minion,
            # but for the master cache, we appended a subdir. Remove it.
            makedirs = True
            dest = str(self.convert_path(cached, cachedir))
            strict = True
        # This is not completely foolproof, but should do the job most
        # of the time and is mostly how the regular client handles it.
        if dest.endswith("/") or self._isdir(dest):
            if not dest.endswith("/"):
                if strict or Path(dest).is_relative_to(
                    Path(self.get_cachedir(cachedir, master=False))
                ):
                    strict = True
                    if not self._rmpath(dest):
                        Path(cached).unlink(missing_ok=True)
                        return False
            if not strict:
                if (
                    url_data.query
                    or len(url_data.path) > 1
                    and not url_data.path.endswith("/")
                ):
                    strpath = url.split("/")[-1]
                else:
                    strpath = "index.html"
                dest = os.path.join(dest, strpath)
        return self._send_file(cached, dest, makedirs, cachedir)

    def get_file(
        self, path, dest="", makedirs=False, saltenv="base", gzip=None, cachedir=None
    ):
        """
        Get a single file from the salt-master
        path must be a salt server location, aka, salt://path/to/file, if
        dest is omitted, then the downloaded file will be placed in the minion
        cache
        """
        src = super().get_file(
            path,
            "",
            makedirs=True,
            saltenv=saltenv,
            cachedir=cachedir,
        )
        if not src:
            return src
        strict = False
        # Passing None evokes the same behavior as an empty string
        # in the parent class as well, which we want to replicate.
        if not dest:
            # The file needs to be cached to the minion cache.
            # We're using the same cachedir on the ssh master and the minion,
            # but for the master cache, we appended a subdir. Remove it.
            makedirs = True
            dest = str(self.convert_path(src, cachedir))
            strict = True

        # This is not completely foolproof, but should do the job most
        # of the time and is mostly how the regular client handles it.
        if dest.endswith("/") or self._isdir(dest):
            if not dest.endswith("/"):
                if strict or Path(dest).is_relative_to(
                    Path(self.get_cachedir(cachedir, master=False))
                ):
                    strict = True
                    if not self._rmpath(dest):
                        Path(src).unlink(missing_ok=True)
                        return ""
            if not strict:
                dest = os.path.join(dest, os.path.basename(src))
        # TODO replicate hash checks to avoid unnecessary transfers,
        # possibly in _send_file to also account for other sources
        return self._send_file(src, dest, makedirs, cachedir)

    def get_template(
        self,
        url,
        dest,
        template="jinja",
        makedirs=False,
        saltenv="base",
        cachedir=None,
        **kwargs,
    ):
        """
        Cache a file then process it as a template
        """
        res = super().get_template(
            url,
            "",
            template=template,
            makedirs=makedirs,
            saltenv=saltenv,
            cachedir=cachedir,
            **kwargs,
        )
        if not res:
            return res
        strict = False
        if not dest:
            # The file needs to be cached to the minion cache.
            # We're using the same cachedir on the ssh master and the minion,
            # but for the master cache, we appended a subdir. Remove it.
            makedirs = True
            dest = str(self.convert_path(res, cachedir))
            strict = True
        if dest.endswith("/") or self._isdir(dest):
            if not dest.endswith("/"):
                if strict or Path(dest).is_relative_to(
                    Path(self.get_cachedir(cachedir, master=False))
                ):
                    strict = True
                    if not self._rmpath(dest):
                        Path(res).unlink(missing_ok=True)
                        return ""
            if not strict:
                dest = os.path.join(dest, os.path.basename(res))
        return self._send_file(res, dest, makedirs, cachedir)

    def _extrn_path(self, url, saltenv, cachedir=None):
        # _extrn_path accesses self.opts["cachedir"] directly,
        # so we have to wrap it here to ensure our master prefix works
        res = super()._extrn_path(url, saltenv, cachedir=cachedir)
        return str(self.convert_path(res, cachedir, master=True))

    def cache_dest(self, url, saltenv="base", cachedir=None):
        """
        Return the expected cache location for the specified URL and
        environment.
        """
        # cache_dest accesses self.opts["cachedir"] directly,
        # so we have to wrap it here to ensure our master prefix works
        res = super().cache_dest(url, saltenv=saltenv, cachedir=cachedir)
        return str(self.convert_path(res, cachedir, master=True))



from __future__ import annotations

import email.policy
import os.path
import re
from email.generator import BytesGenerator
from email.parser import BytesParser

from ..wheelfile import WheelError, WheelFile

DIST_INFO_RE = re.compile(r"^(?P<namever>(?P<name>.+?)-(?P<ver>\d.*?))\.dist-info$")


def pack(directory: str, dest_dir: str, build_number: str | None) -> None:
    """Repack a previously unpacked wheel directory into a new wheel file.

    The .dist-info/WHEEL file must contain one or more tags so that the target
    wheel file name can be determined.

    :param directory: The unpacked wheel directory
    :param dest_dir: Destination directory (defaults to the current directory)
    """
    # Find the .dist-info directory
    dist_info_dirs = [
        fn
        for fn in os.listdir(directory)
        if os.path.isdir(os.path.join(directory, fn)) and DIST_INFO_RE.match(fn)
    ]
    if len(dist_info_dirs) > 1:
        raise WheelError(f"Multiple .dist-info directories found in {directory}")
    elif not dist_info_dirs:
        raise WheelError(f"No .dist-info directories found in {directory}")

    # Determine the target wheel filename
    dist_info_dir = dist_info_dirs[0]
    name_version = DIST_INFO_RE.match(dist_info_dir).group("namever")

    # Read the tags and the existing build number from .dist-info/WHEEL
    wheel_file_path = os.path.join(directory, dist_info_dir, "WHEEL")
    with open(wheel_file_path, "rb") as f:
        info = BytesParser(policy=email.policy.compat32).parse(f)
        tags: list[str] = info.get_all("Tag", [])
        existing_build_number = info.get("Build")

        if not tags:
            raise WheelError(
                f"No tags present in {dist_info_dir}/WHEEL; cannot determine target "
                f"wheel filename"
            )

    # Set the wheel file name and add/replace/remove the Build tag in .dist-info/WHEEL
    build_number = build_number if build_number is not None else existing_build_number
    if build_number is not None:
        del info["Build"]
        if build_number:
            info["Build"] = build_number
            name_version += "-" + build_number

        if build_number != existing_build_number:
            with open(wheel_file_path, "wb") as f:
                BytesGenerator(f, maxheaderlen=0).flatten(info)

    # Reassemble the tags for the wheel file
    tagline = compute_tagline(tags)

    # Repack the wheel
    wheel_path = os.path.join(dest_dir, f"{name_version}-{tagline}.whl")
    with WheelFile(wheel_path, "w") as wf:
        print(f"Repacking wheel as {wheel_path}...", end="", flush=True)
        wf.write_files(directory)

    print("OK")


def compute_tagline(tags: list[str]) -> str:
    """Compute a tagline from a list of tags.

    :param tags: A list of tags
    :return: A tagline
    """
    impls = sorted({tag.split("-")[0] for tag in tags})
    abivers = sorted({tag.split("-")[1] for tag in tags})
    platforms = sorted({tag.split("-")[2] for tag in tags})
    return "-".join([".".join(impls), ".".join(abivers), ".".join(platforms)])



"""
Execution module override for the ``ssh`` resource type.

This module is loaded into the per-type execution-module loader whenever the
``resource_type`` in opts is ``"ssh"``.  It shadows the standard
``salt.modules.cmdmod`` and ``salt.modules.test`` functions for jobs that
are dispatched to SSH resources, delegating the actual work to
:mod:`salt.resources.ssh` via ``__resource_funcs__``.

Because this loader is **only ever used for resource jobs**, there is no need
for the call-time proxy-style guard (``if salt.utils.platform.is_proxy()``).
The managing minion's own jobs continue to use the standard execution modules
loaded in the regular ``self.functions`` loader.

Usage
-----
Any execution module function that should behave differently when targeting
an SSH resource can be implemented here.  Functions not defined in this
module fall through to the standard execution modules in the resource loader.

Example
-------

.. code-block:: bash

    # Ping an SSH resource
    salt -C 'T@ssh:web-01' test.ping

    # Run a shell command on an SSH resource
    salt -C 'T@ssh:web-01' cmd.run 'uptime'
    salt -C 'T@ssh'        cmd.run 'df -h'
"""

import logging

# __resource_funcs__ is injected by the per-type loader at runtime.
# pylint: disable=undefined-variable

log = logging.getLogger(__name__)

# Filename ``cmd.py`` matches the slot name; directory location
# (``salt/resources/ssh/modules/``) gates loading to the ssh per-resource
# loader via the per-type prepend in :func:`salt.loader._module_dirs`.
# No __virtualname__ or __virtual__ gate needed.


# ---------------------------------------------------------------------------
# cmd.* surface
# ---------------------------------------------------------------------------


def run(
    cmd,
    timeout=None,
    **kwargs,
):
    """
    Execute a shell command on the targeted SSH resource and return its
    standard output.

    This is the SSH-resource equivalent of :func:`salt.modules.cmdmod.run`.
    The command is executed directly on the remote host via the SSH Shell
    transport — no Salt thin deployment required.

    :param str cmd: The shell command to run on the remote host.
    :param int timeout: Optional SSH connection timeout in seconds for this
        call.  Overrides the per-resource ``timeout`` configured in Pillar.
    :rtype: str — stdout from the remote command

    CLI Example:

    .. code-block:: bash

        salt -C 'T@ssh:web-01' cmd.run 'uptime'
        salt -C 'T@ssh'        cmd.run 'df -h' timeout=60
    """
    result = __resource_funcs__["ssh.cmd_run"](
        cmd, timeout=timeout
    )  # pylint: disable=undefined-variable
    return result.get("stdout", "")


def run_all(cmd, timeout=None, **kwargs):
    """
    Execute a shell command on the targeted SSH resource and return a dict
    containing ``stdout``, ``stderr``, and ``retcode``.

    This mirrors :func:`salt.modules.cmdmod.run_all` for SSH resources.

    :param str cmd: The shell command to run on the remote host.
    :param int timeout: Optional SSH connection timeout in seconds for this
        call.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt -C 'T@ssh:web-01' cmd.run_all 'uptime'
    """
    return __resource_funcs__["ssh.cmd_run"](
        cmd, timeout=timeout
    )  # pylint: disable=undefined-variable


def retcode(cmd, timeout=None, **kwargs):
    """
    Execute a shell command on the targeted SSH resource and return only the
    exit code.

    :param str cmd: The shell command to run on the remote host.
    :param int timeout: Optional SSH connection timeout in seconds for this
        call.
    :rtype: int

    CLI Example:

    .. code-block:: bash

        salt -C 'T@ssh:web-01' cmd.retcode 'test -f /etc/salt/minion'
    """
    result = __resource_funcs__["ssh.cmd_run"](
        cmd, timeout=timeout
    )  # pylint: disable=undefined-variable
    return result.get("retcode", 1)



"""ZAP Authenticator in a Python Thread.

.. versionadded:: 14.1
"""

# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

import asyncio
from threading import Event, Thread
from typing import Any, List, Optional

import zmq
import zmq.asyncio

from .base import Authenticator


class AuthenticationThread(Thread):
    """A Thread for running a zmq Authenticator

    This is run in the background by ThreadAuthenticator
    """

    pipe: zmq.Socket
    loop: asyncio.AbstractEventLoop
    authenticator: Authenticator
    poller: Optional[zmq.asyncio.Poller] = None

    def __init__(
        self,
        authenticator: Authenticator,
        pipe: zmq.Socket,
    ) -> None:
        super().__init__(daemon=True)
        self.authenticator = authenticator
        self.log = authenticator.log
        self.pipe = pipe

        self.started = Event()

    def run(self) -> None:
        """Start the Authentication Agent thread task"""

        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(self._run())
        finally:
            if self.pipe:
                self.pipe.close()
                self.pipe = None  # type: ignore

            loop.close()

    async def _run(self):
        self.poller = zmq.asyncio.Poller()
        self.poller.register(self.pipe, zmq.POLLIN)
        self.poller.register(self.authenticator.zap_socket, zmq.POLLIN)
        self.started.set()

        while True:
            events = dict(await self.poller.poll())
            if self.pipe in events:
                msg = self.pipe.recv_multipart()
                if self._handle_pipe_message(msg):
                    return
            if self.authenticator.zap_socket in events:
                msg = self.authenticator.zap_socket.recv_multipart()
                await self.authenticator.handle_zap_message(msg)

    def _handle_pipe_message(self, msg: List[bytes]) -> bool:
        command = msg[0]
        self.log.debug("auth received API command %r", command)

        if command == b'TERMINATE':
            return True

        else:
            self.log.error("Invalid auth command from API: %r", command)
            self.pipe.send(b'ERROR')

        return False


class ThreadAuthenticator(Authenticator):
    """Run ZAP authentication in a background thread"""

    pipe: "zmq.Socket"
    pipe_endpoint: str = ''
    thread: AuthenticationThread

    def __init__(
        self,
        context: Optional["zmq.Context"] = None,
        encoding: str = 'utf-8',
        log: Any = None,
    ):
        super().__init__(context=context, encoding=encoding, log=log)
        self.pipe = None  # type: ignore
        self.pipe_endpoint = f"inproc://{id(self)}.inproc"
        self.thread = None  # type: ignore

    def start(self) -> None:
        """Start the authentication thread"""
        # start the Authenticator
        super().start()

        # create a socket pair to communicate with auth thread.
        self.pipe = self.context.socket(zmq.PAIR, socket_class=zmq.Socket)
        self.pipe.linger = 1
        self.pipe.bind(self.pipe_endpoint)
        thread_pipe = self.context.socket(zmq.PAIR, socket_class=zmq.Socket)
        thread_pipe.linger = 1
        thread_pipe.connect(self.pipe_endpoint)
        self.thread = AuthenticationThread(authenticator=self, pipe=thread_pipe)
        self.thread.start()
        if not self.thread.started.wait(timeout=10):
            raise RuntimeError("Authenticator thread failed to start")

    def stop(self) -> None:
        """Stop the authentication thread"""
        if self.pipe:
            self.pipe.send(b'TERMINATE')
            if self.is_alive():
                self.thread.join()
            self.thread = None  # type: ignore
            self.pipe.close()
            self.pipe = None  # type: ignore
        super().stop()

    def is_alive(self) -> bool:
        """Is the ZAP thread currently running?"""
        return bool(self.thread and self.thread.is_alive())

    def __del__(self) -> None:
        self.stop()


__all__ = ['ThreadAuthenticator']



"""
Manage RabbitMQ Users
=====================

Example:

.. code-block:: yaml

    rabbit_user:
      rabbitmq_user.present:
        - password: password
        - force: True
        - tags:
          - monitoring
          - user
        - perms:
          - '/':
            - '.*'
            - '.*'
            - '.*'
        - runas: rabbitmq
"""

import logging

import salt.utils.path
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if RabbitMQ is installed.
    """
    if salt.utils.path.which("rabbitmqctl"):
        return True
    return (False, "Command not found: rabbitmqctl")


def _check_perms_changes(name, newperms, runas=None, existing=None):
    """
    Check whether Rabbitmq user's permissions need to be changed.
    """
    if not newperms:
        return False

    if existing is None:
        try:
            existing = __salt__["rabbitmq.list_user_permissions"](name, runas=runas)
        except CommandExecutionError as err:
            log.error("Error: %s", err)
            return False

    empty_perms = {"configure": "", "write": "", "read": ""}
    perm_need_change = False
    for vhost_perms in newperms:
        for vhost, perms in vhost_perms.items():
            if vhost in existing:
                new_perms = {"configure": perms[0], "write": perms[1], "read": perms[2]}
                existing_vhost = existing[vhost]
                if new_perms != existing_vhost:
                    # This checks for setting permissions to nothing in the state,
                    # when previous state runs have already set permissions to
                    # nothing. We don't want to report a change in this case.
                    if existing_vhost == empty_perms and perms == empty_perms:
                        continue
                    perm_need_change = True
            else:
                perm_need_change = True

    return perm_need_change


def _get_current_tags(name, runas=None):
    """
    Whether Rabbitmq user's tags need to be changed
    """
    try:
        return list(__salt__["rabbitmq.list_users"](runas=runas)[name])
    except CommandExecutionError as err:
        log.error("Error: %s", err)
        return []


def present(name, password=None, force=False, tags=None, perms=(), runas=None):
    """
    Ensure the RabbitMQ user exists.

    name
        User name
    password
        The user's password
    force
        If force is ``True``, the password will be automatically updated without extra password change check.
    tags
        Optional list of tags for the user
    perms
        A list of dicts with vhost keys and 3-tuple values
    runas
        Name of the user to run the command
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    try:
        user = __salt__["rabbitmq.user_exists"](name, runas=runas)
    except CommandExecutionError as err:
        ret["comment"] = f"Error: {err}"
        return ret

    passwd_reqs_update = False
    if user and password is not None:
        try:
            if not __salt__["rabbitmq.check_password"](name, password, runas=runas):
                passwd_reqs_update = True
                log.debug("RabbitMQ user %s password update required", name)
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

    if user and not any((force, perms, tags, passwd_reqs_update)):
        log.debug(
            "RabbitMQ user '%s' exists, password is up to date and force is not set.",
            name,
        )
        ret["comment"] = f"User '{name}' is already present."
        ret["result"] = True
        return ret

    if not user:
        ret["changes"].update({"user": {"old": "", "new": name}})
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"User '{name}' is set to be created."
            return ret

        log.debug("RabbitMQ user '%s' doesn't exist - Creating.", name)
        try:
            __salt__["rabbitmq.add_user"](name, password, runas=runas)
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret
    else:
        log.debug("RabbitMQ user '%s' exists", name)
        if force or passwd_reqs_update:
            if password is not None:
                if not __opts__["test"]:
                    try:
                        __salt__["rabbitmq.change_password"](
                            name, password, runas=runas
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret
                ret["changes"].update({"password": {"old": "", "new": "Set password."}})
            else:
                if not __opts__["test"]:
                    log.debug("Password for %s is not set - Clearing password.", name)
                    try:
                        __salt__["rabbitmq.clear_password"](name, runas=runas)
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret
                ret["changes"].update(
                    {"password": {"old": "Removed password.", "new": ""}}
                )

    if tags is not None:
        current_tags = _get_current_tags(name, runas=runas)
        if isinstance(tags, str):
            tags = tags.split()
        # Diff the tags sets. Symmetric difference operator ^ will give us
        # any element in one set, but not both
        if set(tags) ^ set(current_tags):
            if not __opts__["test"]:
                try:
                    __salt__["rabbitmq.set_user_tags"](name, tags, runas=runas)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret
            ret["changes"].update({"tags": {"old": current_tags, "new": tags}})
    try:
        existing_perms = __salt__["rabbitmq.list_user_permissions"](name, runas=runas)
    except CommandExecutionError as err:
        ret["comment"] = f"Error: {err}"
        return ret

    if _check_perms_changes(name, perms, runas=runas, existing=existing_perms):
        for vhost_perm in perms:
            for vhost, perm in vhost_perm.items():
                if not __opts__["test"]:
                    try:
                        __salt__["rabbitmq.set_permissions"](
                            vhost, name, perm[0], perm[1], perm[2], runas=runas
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret
                new_perms = {
                    vhost: {"configure": perm[0], "write": perm[1], "read": perm[2]}
                }
                if vhost in existing_perms:
                    if existing_perms[vhost] != new_perms[vhost]:
                        if ret["changes"].get("perms") is None:
                            ret["changes"].update({"perms": {"old": {}, "new": {}}})
                        ret["changes"]["perms"]["old"].update(existing_perms[vhost])
                        ret["changes"]["perms"]["new"].update(new_perms)
                else:
                    ret["changes"].update({"perms": {"new": {}}})
                    ret["changes"]["perms"]["new"].update(new_perms)

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = f"'{name}' is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Configuration for '{name}' will change."
        return ret

    ret["comment"] = f"'{name}' was configured."
    return ret


def absent(name, runas=None):
    """
    Ensure the named user is absent

    name
        The name of the user to remove
    runas
        User to run the command
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    try:
        user_exists = __salt__["rabbitmq.user_exists"](name, runas=runas)
    except CommandExecutionError as err:
        ret["comment"] = f"Error: {err}"
        return ret

    if user_exists:
        if not __opts__["test"]:
            try:
                __salt__["rabbitmq.delete_user"](name, runas=runas)
            except CommandExecutionError as err:
                ret["comment"] = f"Error: {err}"
                return ret
        ret["changes"].update({"name": {"old": name, "new": ""}})
    else:
        ret["result"] = True
        ret["comment"] = f"The user '{name}' is not present."
        return ret

    if __opts__["test"] and ret["changes"]:
        ret["result"] = None
        ret["comment"] = f"The user '{name}' will be removed."
        return ret

    ret["result"] = True
    ret["comment"] = f"The user '{name}' was removed."
    return ret



"""
Use pycrypto to generate random passwords on the fly.
"""

import logging
import random
import re
import string

import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError

get_random_bytes = None
try:
    from M2Crypto.Rand import rand_bytes as get_random_bytes
except ImportError:
    try:
        from Cryptodome.Random import get_random_bytes
    except ImportError:
        try:
            from Crypto.Random import get_random_bytes  # nosec
        except ImportError:
            pass
if get_random_bytes is None:
    import secrets

    def get_random_bytes(n):
        return secrets.token_bytes(n)


HAS_RANDOM = True

try:
    import crypt  # pylint: disable=deprecated-module

    # We need to ensure if the right `crypt` is loaded,
    # as LazyLoader can load `salt.utils.crypt` instead of `crypt`
    # if there is Python has no `crypt` (was removed in 3.11).
    HAS_CRYPT = hasattr(crypt, "methods")
except (ImportError, PermissionError):
    HAS_CRYPT = False

try:
    import passlib.context

    HAS_PASSLIB = True
except ImportError:
    HAS_PASSLIB = False

log = logging.getLogger(__name__)


def secure_password(
    length=20,
    use_random=True,
    chars=None,
    lowercase=True,
    uppercase=True,
    digits=True,
    punctuation=True,
    whitespace=False,
    printable=False,
):
    """
    Generate a secure password.
    """
    chars = chars or ""
    if printable:
        # as printable includes all other string character classes
        # the other checks can be skipped
        chars = string.printable
    if not chars:
        if lowercase:
            chars += string.ascii_lowercase
        if uppercase:
            chars += string.ascii_uppercase
        if digits:
            chars += string.digits
        if punctuation:
            chars += string.punctuation
        if whitespace:
            chars += string.whitespace
    try:
        length = int(length)
        pw = ""
        while len(pw) < length:
            if HAS_RANDOM and use_random:
                encoding = None
                if salt.utils.platform.is_windows():
                    encoding = "UTF-8"
                while True:
                    try:
                        char = salt.utils.stringutils.to_str(
                            get_random_bytes(1), encoding=encoding
                        )
                        break
                    except UnicodeDecodeError:
                        continue
                pw += re.sub(
                    salt.utils.stringutils.to_str(
                        rf"[^{re.escape(chars)}]", encoding=encoding
                    ),
                    "",
                    char,
                )
            else:
                pw += random.SystemRandom().choice(chars)
        return pw
    except Exception as exc:  # pylint: disable=broad-except
        log.exception("Failed to generate secure passsword")
        raise CommandExecutionError(str(exc))


if HAS_CRYPT:
    methods = {m.name.lower(): m for m in crypt.methods}
else:
    methods = {}
known_methods = ["sha512", "sha256", "blowfish", "md5", "crypt"]


def _gen_hash_passlib(crypt_salt=None, password=None, algorithm=None):
    """
    Generate a /etc/shadow-compatible hash for a non-local system
    """
    # these are the passlib equivalents to the 'known_methods' defined in crypt
    schemes = ["sha512_crypt", "sha256_crypt", "bcrypt", "md5_crypt", "des_crypt"]

    ctx = passlib.context.CryptContext(schemes=schemes)

    kwargs = {"secret": password, "scheme": schemes[known_methods.index(algorithm)]}
    if crypt_salt and "$" in crypt_salt:
        # this salt has a rounds specifier.
        #  passlib takes it as a separate parameter, split it out
        roundsstr, split_salt = crypt_salt.split("$")
        rounds = int(roundsstr.split("=")[-1])
        kwargs.update({"salt": split_salt, "rounds": rounds})
    else:
        # relaxed = allow salts that are too long
        kwargs.update({"salt": crypt_salt, "relaxed": True})
    return ctx.hash(**kwargs)


def _gen_hash_crypt(crypt_salt=None, password=None, algorithm=None):
    """
    Generate /etc/shadow hash using the native crypt module
    """
    if crypt_salt is None:
        # setting crypt_salt to the algorithm makes crypt generate
        #  a salt compatible with the specified algorithm.
        crypt_salt = methods[algorithm]
    else:
        if algorithm != "crypt":
            # all non-crypt algorithms are specified as part of the salt
            crypt_salt = f"${methods[algorithm].ident}${crypt_salt}"

    try:
        ret = crypt.crypt(password, crypt_salt)
    except OSError:
        ret = None
    return ret


def gen_hash(crypt_salt=None, password=None, algorithm=None):
    """
    Generate /etc/shadow hash
    """
    if password is None:
        password = secure_password()

    if algorithm is None:
        # prefer the most secure natively supported method
        algorithm = crypt.methods[0].name.lower() if HAS_CRYPT else known_methods[0]

    if algorithm == "crypt" and crypt_salt and len(crypt_salt) != 2:
        log.warning("Hash salt is too long for 'crypt' hash.")

    if HAS_CRYPT and algorithm in methods:
        return _gen_hash_crypt(
            crypt_salt=crypt_salt, password=password, algorithm=algorithm
        )
    elif HAS_PASSLIB and algorithm in known_methods:
        return _gen_hash_passlib(
            crypt_salt=crypt_salt, password=password, algorithm=algorithm
        )
    else:
        raise SaltInvocationError(
            "Cannot hash using '{}' hash algorithm. Natively supported "
            "algorithms are: {}. If passlib is installed ({}), the supported "
            "algorithms are: {}.".format(
                algorithm, list(methods), HAS_PASSLIB, known_methods
            )
        )



"""
Helpers for :mod:`tarfile` across Python versions.

Python 3.12 added optional extraction filters (PEP 706). Passing
``filter='data'`` limits surprising or unsafe features when unpacking
archives from less trusted sources.
"""

from __future__ import annotations

import sys
import tarfile


def extractall(
    tar: tarfile.TarFile,
    path: str = "",
    members=None,
    *,
    numeric_owner: bool = False,
) -> None:
    """
    Like :meth:`tarfile.TarFile.extractall` but use ``filter='data'`` on
    Python 3.12+ when supported.
    """
    if sys.version_info >= (3, 12) and isinstance(tar, tarfile.TarFile):
        tar.extractall(  # nosec B202
            path,
            members,
            numeric_owner=numeric_owner,
            filter="data",
        )
    elif members is not None or numeric_owner:
        tar.extractall(path, members, numeric_owner=numeric_owner)  # nosec B202
    else:
        tar.extractall(path)  # nosec B202


def extract(
    tar: tarfile.TarFile,
    member,
    path: str = "",
    *,
    set_attrs: bool = True,
    numeric_owner: bool = False,
) -> None:
    """
    Like :meth:`tarfile.TarFile.extract` but use ``filter='data'`` on
    Python 3.12+ when supported.
    """
    if sys.version_info >= (3, 12) and isinstance(tar, tarfile.TarFile):
        tar.extract(  # nosec B202
            member,
            path,
            set_attrs=set_attrs,
            numeric_owner=numeric_owner,
            filter="data",
        )
    elif not set_attrs or numeric_owner:
        tar.extract(  # nosec B202
            member,
            path,
            set_attrs=set_attrs,
            numeric_owner=numeric_owner,
        )
    else:
        tar.extract(member, path)  # nosec B202



"""
Manage account locks on AIX systems

.. versionadded:: 2018.3.0

:depends: none
"""

# Import python librarie
import logging

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "shadow"


def __virtual__():
    """
    Only load if kernel is AIX
    """
    if __grains__["kernel"] == "AIX":
        return __virtualname__
    return (
        False,
        "The aix_shadow execution module failed to load: "
        "only available on AIX systems.",
    )


def login_failures(user):
    """
    Query for all accounts which have 3 or more login failures.

    CLI Example:

    .. code-block:: bash

        salt <minion_id> shadow.login_failures ALL
    """

    cmd = f"lsuser -a unsuccessful_login_count {user}"
    cmd += " | grep -E 'unsuccessful_login_count=([3-9]|[0-9][0-9]+)'"
    out = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=True)

    ret = []

    lines = out["stdout"].splitlines()
    for line in lines:
        ret.append(line.split()[0])

    return ret


def locked(user):
    """
    Query for all accounts which are flagged as locked.

    CLI Example:

    .. code-block:: bash

        salt <minion_id> shadow.locked ALL
    """

    cmd = f"lsuser -a account_locked {user}"
    cmd += ' | grep "account_locked=true"'
    out = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=True)

    ret = []

    lines = out["stdout"].splitlines()
    for line in lines:
        ret.append(line.split()[0])

    return ret


def unlock(user):
    """
    Unlock user for locked account

    CLI Example:

    .. code-block:: bash

        salt <minion_id> shadow.unlock user
    """

    cmd = (
        "chuser account_locked=false {0} | "
        'chsec -f /etc/security/lastlog -a "unsuccessful_login_count=0" -s {0}'.format(
            user
        )
    )
    ret = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=True)

    return ret



# Copyright 2022-2026 Broadcom.
# SPDX-License-Identifier: Apache-2.0
"""
A script to ensure the proper rpaths are in place for the relenv environment.
"""

from __future__ import annotations

import logging
import os as _os
import pathlib
import shutil as _shutil
import subprocess as _subprocess
import sys as _sys

log = logging.getLogger(__name__)

os = _os
shutil = _shutil
subprocess = _subprocess
sys = _sys

__all__ = [
    "is_macho",
    "is_elf",
    "parse_otool_l",
    "parse_readelf_d",
    "parse_macho",
    "parse_rpath",
    "handle_macho",
    "is_in_dir",
    "handle_elf",
    "patch_rpath",
    "main",
    "os",
    "shutil",
    "subprocess",
]


LIBCLIBS = [
    "linux-vdso.so.1",
    "libc.so.6",
    "librt.so.1",
    "libm.so.6",
    "libmd.so.0",
    "libpthread.so.0",
    # Adjust rpath of libcrypt because we include libxcrypt
    # "libcrypt.so.1",
    "libdl.so.2",
    "libmemusage.so",
    "libnsl.so.1",
    "libnss_compat.so.2",
    "libnss_db.so.2",
    "libnss_dns.so.2",
    "libnss_files.so.2",
    "libnss_hesiod.so.2",
    "libpcprofile.so.2",
    "libresolv.so.2",
    "librt.so.1",
    "libthread_db.so.1",
    "libutil.so.1",
    "libutil.so.2",
    # libgcc is not technically glibc but shares the same kind of backwards
    # compatablity guarantees.
    "libgcc_s.so.2",
    "libgcc_s.so.1",
]

LC_ID_DYLIB = "LC_ID_DYLIB"
LC_LOAD_DYLIB = "LC_LOAD_DYLIB"
LC_RPATH = "LC_RPATH"

# Cache for readelf binary path
_READELF_BINARY: str | None = None

# Cache for patchelf binary path
_PATCHELF_BINARY: str | None = None


def _get_readelf_binary() -> str:
    """
    Get the path to readelf binary, preferring toolchain version.

    Returns the cached value if already computed. On Linux, prefers the
    toolchain's readelf over the system version. Falls back to "readelf"
    from PATH if toolchain is unavailable.

    :return: Path to readelf binary
    :rtype: str
    """
    global _READELF_BINARY
    if _READELF_BINARY is not None:
        return _READELF_BINARY

    # Only Linux has the toolchain with readelf
    if sys.platform == "linux":
        try:
            from relenv.common import get_toolchain, get_triplet

            toolchain = get_toolchain()
            if toolchain:
                triplet = get_triplet()
                toolchain_readelf = toolchain / "bin" / f"{triplet}-readelf"
                if toolchain_readelf.exists():
                    _READELF_BINARY = str(toolchain_readelf)
                    return _READELF_BINARY
        except Exception:
            # Fall through to system readelf
            pass

    # Fall back to system readelf
    _READELF_BINARY = "readelf"
    return _READELF_BINARY


def _get_patchelf_binary() -> str:
    """
    Get the path to patchelf binary, preferring toolchain version.

    Returns the cached value if already computed. On Linux, prefers the
    toolchain's patchelf over the system version. Falls back to "patchelf"
    from PATH if toolchain is unavailable.

    :return: Path to patchelf binary
    :rtype: str
    """
    global _PATCHELF_BINARY
    if _PATCHELF_BINARY is not None:
        return _PATCHELF_BINARY

    # Only Linux has the toolchain with patchelf
    if sys.platform == "linux":
        try:
            from relenv.common import get_toolchain

            toolchain = get_toolchain()
            if toolchain:
                toolchain_patchelf = toolchain / "bin" / "patchelf"
                if toolchain_patchelf.exists():
                    _PATCHELF_BINARY = str(toolchain_patchelf)
                    return _PATCHELF_BINARY
        except Exception:
            # Fall through to system patchelf
            pass

    # Fall back to system patchelf
    _PATCHELF_BINARY = "patchelf"
    return _PATCHELF_BINARY


def is_macho(path: str | os.PathLike[str]) -> bool:
    """
    Determines whether the given file is a macho file.

    :param path: The path to the file to check
    :type path: str

    :return: Whether the file is a macho file
    :rtype: bool
    """
    with open(path, "rb") as fp:
        magic = fp.read(4)
    # XXX: Handle 64bit, 32bit, ppc, arm
    return magic in [b"\xcf\xfa\xed\xfe"]


def is_elf(path: str | os.PathLike[str]) -> bool:
    """
    Determines whether the given file is an ELF file.

    :param path: The path to the file to check
    :type path: str

    :return: Whether the file is an ELF file
    :rtype: bool
    """
    with open(path, "rb") as fp:
        magic = fp.read(4)
    return magic == b"\x7f\x45\x4c\x46"


def parse_otool_l(stdout: str) -> dict[str, list[str]]:
    """
    Parse the output of ``otool -l <path>``.

    :param stdout: The output of the ``otool -l <path>`` command
    :type stdout: str

    :return: The parsed relevant output with command keys and path values
    :rtype: dict
    """
    in_cmd = False
    cmd: str | None = None
    name: str | None = None
    data: dict[str, list[str]] = {}
    for line in [x.strip() for x in stdout.split("\n")]:
        if not line:
            continue

        if line.split()[0] == "cmd":
            in_cmd = False
            if cmd is not None and name is not None:
                data.setdefault(cmd, []).append(name)
                cmd = None
                name = None
            command = line.split()[-1]
            if command in (LC_ID_DYLIB, LC_LOAD_DYLIB, "LC_RPATH"):
                cmd = command
                in_cmd = True

        if in_cmd:
            parts = line.split()
            if parts[0] == "name" and len(parts) > 1:
                name = parts[1]
            if parts[0] == "path" and len(parts) > 1:
                name = parts[1]

    if in_cmd and cmd is not None and name is not None:
        data.setdefault(cmd, []).append(name)

    return data


def parse_readelf_d(stdout: str) -> list[str]:
    """
    Parse the output of ``readelf -d <path>``.

    :param stdout: The output of the ``readelf -d <path>`` command
    :type stdout: str

    :return: The RPATH values
    :rtype: list
    """
    for line in stdout.splitlines():
        # Find either RPATH or READPATH
        if line.find("PATH") == -1:
            continue
        return line.split(":", 1)[-1].strip().strip("[").strip("]").split(":")
    return []


def parse_macho(path: str | os.PathLike[str]) -> dict[str, list[str]] | None:
    """
    Run ``otool -l <path>`` and return its parsed output.

    :param path: The path to the file
    :type path: str

    :return: The parsed relevant RPATH content, or None if it isn't an object file
    :rtype: dict or None
    """
    proc = subprocess.run(["otool", "-l", path], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    stdout = proc.stdout.decode()
    if stdout.find("is not an object file") != -1:
        return None
    return parse_otool_l(stdout)


def parse_rpath(path: str | os.PathLike[str]) -> list[str]:
    """
    Run ``readelf -d <path>`` and return its parsed output.

    :param path: The path to the file
    :type path: str

    :return: The RPATH's found.
    :rtype: list
    """
    readelf = _get_readelf_binary()
    proc = subprocess.run([readelf, "-d", path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return parse_readelf_d(proc.stdout.decode())


def handle_macho(
    path: str | os.PathLike[str],
    root_dir: str | os.PathLike[str],
    rpath_only: bool,
) -> dict[str, list[str]] | None:
    """
    Ensure the given macho file has the correct rpath and is in th correct location.

    :param path: The path to a macho file
    :type path: str
    :param root_dir: The directory the file needs to reside under
    :type root_dir: str
    :param rpath_only: If true, only ensure the correct rpaths are present and don't copy the file
    :type rpath_only: bool

    :return: The information from ``parse_macho`` on the macho file.
    """
    path_obj = pathlib.Path(path)
    obj = parse_macho(path_obj)
    path_str = str(path_obj)
    log.info("Processing file %s %r", path_str, obj)
    if not obj:
        return None
    if LC_LOAD_DYLIB in obj:
        for x in obj[LC_LOAD_DYLIB]:
            if x.startswith("@"):
                log.info("Skipping dynamic load: %s", x)
                continue
            if os.path.exists(x):
                target_dir = pathlib.Path(root_dir).resolve()
                y = target_dir / os.path.basename(x)
                if not os.path.exists(y):
                    if rpath_only:
                        log.warning("In `rpath_only mode` but %s is not in %s", x, y)
                        continue
                    else:
                        shutil.copy(x, y)
                        shutil.copymode(x, y)
                        log.info("Copied %s to %s", x, y)
                log.info("Use %s to %s", y, path_str)
                z = pathlib.Path("@loader_path") / os.path.relpath(y, path_obj.resolve().parent)
                cmd = ["install_name_tool", "-change", x, str(z), path_str]
                subprocess.run(cmd)
                log.info("Changed %s to %s in %s", x, z, path_str)
    return obj


def is_in_dir(filepath: str | os.PathLike[str], directory: str | os.PathLike[str]) -> bool:
    """
    Determines whether a file is contained within a directory.

    :param filepath: The path to the file to check
    :type filepath: str
    :param directory: The directory to check within
    :type directory: str

    :return: Whether the file is contained within the given directory
    :rtype: bool
    """
    return os.path.realpath(filepath).startswith(os.path.realpath(directory) + os.sep)


def remove_rpath(path: str | os.PathLike[str]) -> bool:
    """
    Remove the rpath from a given ELF file.

    :param path: The path to an ELF file
    :type path: str

    :return: True if successful, False otherwise
    :rtype: bool
    """
    old_rpath = parse_rpath(path)
    if not old_rpath:
        # No RPATH to remove
        return True

    log.info("Remove RPATH from %s (was: %s)", path, old_rpath)
    patchelf = _get_patchelf_binary()
    proc = subprocess.run(
        [patchelf, "--remove-rpath", path],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    if proc.returncode:
        log.error("Failed to remove RPATH from %s: %s", path, proc.stderr.decode())
        return False
    return True


def patch_rpath(
    path: str | os.PathLike[str],
    new_rpath: str,
    only_relative: bool = True,
) -> str | bool:
    """
    Patch the rpath of a given ELF file.

    :param path: The path to an ELF file
    :type path: str
    :param new_rpath: The new rpath to add
    :type new_rpath: str
    :param only_relative: Whether or not to remove non-relative rpaths, defaults to True
    :type only_relative: bool, optional

    :return: The new rpath, or False if patching failed
    :rtype: str or bool
    """
    old_rpath = parse_rpath(path)

    # Remove non-relative rpaths if needed
    if only_relative:
        old_rpath = [rpath for rpath in old_rpath if rpath.startswith("$ORIGIN")]

    if new_rpath not in old_rpath:
        patched_rpath = ":".join([new_rpath] + old_rpath)
        log.info("Set RPATH=%s %s", patched_rpath, path)
        patchelf = _get_patchelf_binary()
        proc = subprocess.run(
            [patchelf, "--force-rpath", "--set-rpath", patched_rpath, path],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        if proc.returncode:
            return False
        return patched_rpath
    return ":".join(old_rpath)


def handle_elf(
    path: str | os.PathLike[str],
    libs: str | os.PathLike[str],
    rpath_only: bool,
    root: str | os.PathLike[str] | None = None,
) -> None:
    """
    Handle the parsing and pathcing of an ELF file.

    :param path: The path of the ELF file
    :type path: str
    :param libs: The libs directory
    :type libs: str
    :param rpath_only: If true, only ensure the correct rpaths are present and don't copy the file
    :type rpath_only: bool
    :param root: The directory to ensure the file is under, defaults to None
    :type root: str, optional
    """
    if root is None:
        root = libs
    proc = subprocess.run(["ldd", path], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    needs_rpath = False

    for line in proc.stdout.decode().splitlines():
        if line.find("=>") == -1:
            log.debug("Skip ldd output line: %s", line)
            continue

        lib_name, location_info = [_.strip() for _ in line.split("=>", 1)]
        if location_info == "not found":
            # It is likely that something was not compiled correctly
            log.warning("Unable to find library %s linked from %s", lib_name, path)
            continue

        linked_lib = location_info.rsplit(" ", 1)[0].strip()
        lib_basename = os.path.basename(linked_lib)

        if lib_name in LIBCLIBS:
            log.debug("Skipping glibc lib %s", lib_name)
            continue

        if is_in_dir(linked_lib, root):
            needs_rpath = True
            log.info("File already within root directory: %s", linked_lib)
            continue

        relocated_path = os.path.join(libs, lib_basename)

        if os.path.exists(relocated_path):
            log.debug("Relocated library exists: %s", relocated_path)
        elif rpath_only:
            log.warning("In `rpath_only mode` but %s is not in %s", linked_lib, root)
        else:
            # If we aren't in `rpath_only` mode, we can copy
            log.info("Copy %s to %s", linked_lib, relocated_path)
            shutil.copy(linked_lib, relocated_path)
            shutil.copymode(linked_lib, relocated_path)
            needs_rpath = True

    if needs_rpath:
        relpart = os.path.relpath(libs, os.path.dirname(path))
        if relpart == ".":
            relpath = "$ORIGIN"
        else:
            relpath = str(pathlib.Path("$ORIGIN") / relpart)

        log.info("Adjust rpath of %s to %s", path, relpath)
        patch_rpath(path, relpath)
    else:
        # No relenv libraries are linked, so RPATH is not needed
        # Remove any existing RPATH to avoid security/correctness issues
        remove_rpath(path)


def main(
    root: str | os.PathLike[str],
    libs_dir: str | os.PathLike[str] | None = None,
    rpath_only: bool = True,
    log_level: str = "DEBUG",
    log_file_name: str = "<stdout>",
) -> None:
    """
    The entrypoint into the relocate script.

    :param root: The root directory to operate traverse for files to be patched
    :type root: str
    :param libs_dir: The directory to place the libraries in, defaults to None
    :type libs_dir: str, optional
    :param rpath_only: If true, only ensure the correct rpaths are present and don't copy the file, defaults to True
    :type rpath_only: bool, optional
    :param log_level: The level to log at, defaults to "INFO"
    :type log_level: str, optional
    """
    level = logging.getLevelName(log_level.upper())
    if log_file_name != "<stdout>":
        logging.basicConfig(
            level=level,
            format="%(asctime)s %(message)s",
            filename=log_file_name,
            filemode="w",
        )
    else:
        logging.basicConfig(
            level=level,
            format="%(asctime)s %(message)s",
        )
    root_dir = str(pathlib.Path(root).resolve())
    if libs_dir is None:
        libs_dir = pathlib.Path(root_dir, "lib")
    libs_dir = str(pathlib.Path(libs_dir).resolve())
    processed: dict[str, dict[str, list[str]] | None] = {}
    found = True
    while found:
        found = False
        for root, dirs, files in os.walk(root_dir):
            for file in files:
                path = os.path.join(root, file)
                if path in processed:
                    continue
                log.debug("Checking %s", path)

                if is_macho(path):
                    log.info("Found Mach-O %s", path)
                    _ = handle_macho(path, libs_dir, rpath_only)
                    if _ is not None:
                        processed[path] = _
                        found = True
                elif is_elf(path):
                    log.info("Found ELF %s", path)
                    handle_elf(path, libs_dir, rpath_only, root_dir)


if __name__ == "__main__":
    import sys

    if not hasattr(sys, "RELENV"):
        raise RuntimeError("Not in a relenv environment")
    main(sys.RELENV)



import copy
from threading import Lock
from typing import Dict, Iterable, List, Optional, Protocol

from .metrics_core import Metric


class Collector(Protocol):
    def collect(self) -> Iterable[Metric]:
        """Collect metrics."""


class _EmptyCollector:
    def collect(self) -> Iterable[Metric]:
        return []


class CollectorRegistry:
    """Metric collector registry.

    Collectors must have a no-argument method 'collect' that returns a list of
    Metric objects. The returned metrics should be consistent with the Prometheus
    exposition formats.
    """

    def __init__(self, auto_describe: bool = False, target_info: Optional[Dict[str, str]] = None,
                 support_collectors_without_names: bool = False):
        self._collector_to_names: Dict[Collector, List[str]] = {}
        self._names_to_collectors: Dict[str, Collector] = {}
        self._auto_describe = auto_describe
        self._lock = Lock()
        self._target_info: Optional[Dict[str, str]] = {}
        self._support_collectors_without_names = support_collectors_without_names
        self._collectors_without_names: List[Collector] = []
        self.set_target_info(target_info)

    def register(self, collector: Collector) -> None:
        """Add a collector to the registry."""
        with self._lock:
            names = self._get_names(collector)
            duplicates = set(self._names_to_collectors).intersection(names)
            if duplicates:
                raise ValueError(
                    'Duplicated timeseries in CollectorRegistry: {}'.format(
                        duplicates))
            for name in names:
                self._names_to_collectors[name] = collector
            self._collector_to_names[collector] = names
            if self._support_collectors_without_names and not names:
                self._collectors_without_names.append(collector)

    def unregister(self, collector: Collector) -> None:
        """Remove a collector from the registry."""
        with self._lock:
            for name in self._collector_to_names[collector]:
                del self._names_to_collectors[name]
            del self._collector_to_names[collector]

    def _get_names(self, collector):
        """Get names of timeseries the collector produces and clashes with."""
        desc_func = None
        # If there's a describe function, use it.
        try:
            desc_func = collector.describe
        except AttributeError:
            pass
        # Otherwise, if auto describe is enabled use the collect function.
        if not desc_func and self._auto_describe:
            desc_func = collector.collect

        if not desc_func:
            return []

        result = []
        type_suffixes = {
            'counter': ['_total', '_created'],
            'summary': ['_sum', '_count', '_created'],
            'histogram': ['_bucket', '_sum', '_count', '_created'],
            'gaugehistogram': ['_bucket', '_gsum', '_gcount'],
            'info': ['_info'],
        }
        for metric in desc_func():
            result.append(metric.name)
            for suffix in type_suffixes.get(metric.type, []):
                result.append(metric.name + suffix)
        return result

    def collect(self) -> Iterable[Metric]:
        """Yields metrics from the collectors in the registry."""
        collectors = None
        ti = None
        with self._lock:
            collectors = copy.copy(self._collector_to_names)
            if self._target_info:
                ti = self._target_info_metric()
        if ti:
            yield ti
        for collector in collectors:
            yield from collector.collect()

    def restricted_registry(self, names: Iterable[str]) -> "RestrictedRegistry":
        """Returns object that only collects some metrics.

        Returns an object which upon collect() will return
        only samples with the given names.

        Intended usage is:
            generate_latest(REGISTRY.restricted_registry(['a_timeseries']), escaping)

        Experimental."""
        names = set(names)
        return RestrictedRegistry(names, self)

    def set_target_info(self, labels: Optional[Dict[str, str]]) -> None:
        with self._lock:
            if labels:
                if not self._target_info and 'target_info' in self._names_to_collectors:
                    raise ValueError('CollectorRegistry already contains a target_info metric')
                self._names_to_collectors['target_info'] = _EmptyCollector()
            elif self._target_info:
                self._names_to_collectors.pop('target_info', None)
            self._target_info = labels

    def get_target_info(self) -> Optional[Dict[str, str]]:
        with self._lock:
            return self._target_info

    def _target_info_metric(self):
        m = Metric('target', 'Target metadata', 'info')
        m.add_sample('target_info', self._target_info, 1)
        return m

    def get_sample_value(self, name: str, labels: Optional[Dict[str, str]] = None) -> Optional[float]:
        """Returns the sample value, or None if not found.

        This is inefficient, and intended only for use in unittests.
        """
        if labels is None:
            labels = {}
        for metric in self.collect():
            for s in metric.samples:
                if s.name == name and s.labels == labels:
                    return s.value
        return None


class RestrictedRegistry:
    def __init__(self, names: Iterable[str], registry: CollectorRegistry):
        self._name_set = set(names)
        self._registry = registry

    def collect(self) -> Iterable[Metric]:
        collectors = set(self._registry._collectors_without_names)
        target_info_metric = None
        with self._registry._lock:
            if 'target_info' in self._name_set and self._registry._target_info:
                target_info_metric = self._registry._target_info_metric()
            for name in self._name_set:
                if name != 'target_info' and name in self._registry._names_to_collectors:
                    collectors.add(self._registry._names_to_collectors[name])
        if target_info_metric:
            yield target_info_metric
        for collector in collectors:
            for metric in collector.collect():
                m = metric._restricted_metric(self._name_set)
                if m:
                    yield m


REGISTRY = CollectorRegistry(auto_describe=True)



"""
Get Version information from Windows
"""

# http://stackoverflow.com/questions/32300004/python-ctypes-getting-0-with-getversionex-function

import ctypes

HAS_WIN32 = True
try:
    from ctypes.wintypes import BYTE, DWORD, WCHAR, WORD

    import win32net
    import win32netcon
except (ImportError, ValueError):
    HAS_WIN32 = False

if HAS_WIN32:
    kernel32 = ctypes.WinDLL(
        "kernel32",
        use_last_error=True,
    )


# Although utils are often directly imported, it is also possible to use the
# loader.
def __virtual__():
    """
    Only load if Win32 Libraries are installed
    """
    if not HAS_WIN32:
        return False, "This utility requires pywin32"

    return "win_osinfo"


def os_version_info_ex():
    """
    Helper function to return the results of the GetVersionExW Windows API call.
    It is a ctypes Structure that contains Windows OS Version information.

    Returns:
        class: An instance of a class containing version info
    """
    if not HAS_WIN32:
        return

    class OSVersionInfo(ctypes.Structure):
        _fields_ = (
            ("dwOSVersionInfoSize", DWORD),
            ("dwMajorVersion", DWORD),
            ("dwMinorVersion", DWORD),
            ("dwBuildNumber", DWORD),
            ("dwPlatformId", DWORD),
            ("szCSDVersion", WCHAR * 128),
        )

        def __init__(self, *args, **kwds):
            super().__init__(*args, **kwds)
            self.dwOSVersionInfoSize = ctypes.sizeof(self)
            kernel32.GetVersionExW(ctypes.byref(self))

    class OSVersionInfoEx(OSVersionInfo):
        _fields_ = (
            ("wServicePackMajor", WORD),
            ("wServicePackMinor", WORD),
            ("wSuiteMask", WORD),
            ("wProductType", BYTE),
            ("wReserved", BYTE),
        )

    return OSVersionInfoEx()


def get_os_version_info():
    info = os_version_info_ex()
    ret = {
        "MajorVersion": info.dwMajorVersion,
        "MinorVersion": info.dwMinorVersion,
        "BuildNumber": info.dwBuildNumber,
        "PlatformID": info.dwPlatformId,
        "ServicePackMajor": info.wServicePackMajor,
        "ServicePackMinor": info.wServicePackMinor,
        "SuiteMask": info.wSuiteMask,
        "ProductType": info.wProductType,
    }

    return ret


def get_join_info():
    """
    Gets information about the domain/workgroup. This will tell you if the
    system is joined to a domain or a workgroup

    .. versionadded:: 2018.3.4

    Returns:
        dict: A dictionary containing the domain/workgroup and its status
    """
    info = win32net.NetGetJoinInformation()
    status = {
        win32netcon.NetSetupUnknown: "Unknown",
        win32netcon.NetSetupUnjoined: "Unjoined",
        win32netcon.NetSetupWorkgroupName: "Workgroup",
        win32netcon.NetSetupDomainName: "Domain",
    }
    return {"Domain": info[0], "DomainType": status[info[1]]}



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

DESCRIPTOR: _descriptor.FileDescriptor
FIELD_INFO_FIELD_NUMBER: _ClassVar[int]
field_info: _descriptor.FieldDescriptor

class FieldInfo(_message.Message):
    __slots__ = ("format", "referenced_types")
    class Format(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        FORMAT_UNSPECIFIED: _ClassVar[FieldInfo.Format]
        UUID4: _ClassVar[FieldInfo.Format]
        IPV4: _ClassVar[FieldInfo.Format]
        IPV6: _ClassVar[FieldInfo.Format]
        IPV4_OR_IPV6: _ClassVar[FieldInfo.Format]

    FORMAT_UNSPECIFIED: FieldInfo.Format
    UUID4: FieldInfo.Format
    IPV4: FieldInfo.Format
    IPV6: FieldInfo.Format
    IPV4_OR_IPV6: FieldInfo.Format
    FORMAT_FIELD_NUMBER: _ClassVar[int]
    REFERENCED_TYPES_FIELD_NUMBER: _ClassVar[int]
    format: FieldInfo.Format
    referenced_types: _containers.RepeatedCompositeFieldContainer[TypeReference]
    def __init__(
        self,
        format: _Optional[_Union[FieldInfo.Format, str]] = ...,
        referenced_types: _Optional[_Iterable[_Union[TypeReference, _Mapping]]] = ...,
    ) -> None: ...

class TypeReference(_message.Message):
    __slots__ = ("type_name",)
    TYPE_NAME_FIELD_NUMBER: _ClassVar[int]
    type_name: str
    def __init__(self, type_name: _Optional[str] = ...) -> None: ...



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

DEVICE_ID: Final = "device.id"
"""
A unique identifier representing the device.
Note: Its value SHOULD be identical for all apps on a device and it SHOULD NOT change if an app is uninstalled and re-installed.
However, it might be resettable by the user for all apps on a device.
Hardware IDs (e.g. vendor-specific serial number, IMEI or MAC address) MAY be used as values.

More information about Android identifier best practices can be found in the [Android user data IDs guide](https://developer.android.com/training/articles/user-data-ids).

> [!WARNING]
>
> This attribute may contain sensitive (PII) information. Caution should be taken when storing personal data or anything which can identify a user. GDPR and data protection laws may apply,
> ensure you do your own due diligence.
>
> Due to these reasons, this identifier is not recommended for consumer applications and will likely result in rejection from both Google Play and App Store.
> However, it may be appropriate for specific enterprise scenarios, such as kiosk devices or enterprise-managed devices, with appropriate compliance clearance.
> Any instrumentation providing this identifier MUST implement it as an opt-in feature.
>
> See [`app.installation.id`](/docs/registry/attributes/app.md#app-installation-id) for a more privacy-preserving alternative.
"""

DEVICE_MANUFACTURER: Final = "device.manufacturer"
"""
The name of the device manufacturer.
Note: The Android OS provides this field via [Build](https://developer.android.com/reference/android/os/Build#MANUFACTURER). iOS apps SHOULD hardcode the value `Apple`.
"""

DEVICE_MODEL_IDENTIFIER: Final = "device.model.identifier"
"""
The model identifier for the device.
Note: It's recommended this value represents a machine-readable version of the model identifier rather than the market or consumer-friendly name of the device.
"""

DEVICE_MODEL_NAME: Final = "device.model.name"
"""
The marketing name for the device model.
Note: It's recommended this value represents a human-readable version of the device model rather than a machine-readable alternative.
"""



"""
Return data to a postgresql server

.. note::
    There are three PostgreSQL returners.  Any can function as an external
    :ref:`master job cache <external-job-cache>`. but each has different
    features.  SaltStack recommends
    :mod:`returners.pgjsonb <salt.returners.pgjsonb>` if you are working with
    a version of PostgreSQL that has the appropriate native binary JSON types.
    Otherwise, review
    :mod:`returners.postgres <salt.returners.postgres>` and
    :mod:`returners.postgres_local_cache <salt.returners.postgres_local_cache>`
    to see which module best suits your particular needs.

:maintainer:    None
:maturity:      New
:depends:       psycopg2
:platform:      all

To enable this returner the minion will need the psycopg2 installed and
the following values configured in the minion or master config:

.. code-block:: yaml

    returner.postgres.host: 'salt'
    returner.postgres.user: 'salt'
    returner.postgres.passwd: 'salt'
    returner.postgres.db: 'salt'
    returner.postgres.port: 5432

Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location:

.. code-block:: yaml

    alternative.returner.postgres.host: 'salt'
    alternative.returner.postgres.user: 'salt'
    alternative.returner.postgres.passwd: 'salt'
    alternative.returner.postgres.db: 'salt'
    alternative.returner.postgres.port: 5432

Running the following commands as the postgres user should create the database
correctly:

.. code-block:: sql

    psql << EOF
    CREATE ROLE salt WITH PASSWORD 'salt';
    CREATE DATABASE salt WITH OWNER salt;
    EOF

    psql -h localhost -U salt << EOF
    --
    -- Table structure for table 'jids'
    --

    DROP TABLE IF EXISTS jids;
    CREATE TABLE jids (
      jid   varchar(20) PRIMARY KEY,
      load  text NOT NULL
    );

    --
    -- Table structure for table 'salt_returns'
    --

    DROP TABLE IF EXISTS salt_returns;
    CREATE TABLE salt_returns (
      fun       varchar(50) NOT NULL,
      jid       varchar(255) NOT NULL,
      return    text NOT NULL,
      full_ret  text,
      id        varchar(255) NOT NULL,
      success   varchar(10) NOT NULL,
      alter_time   TIMESTAMP WITH TIME ZONE DEFAULT now()
    );

    CREATE INDEX idx_salt_returns_id ON salt_returns (id);
    CREATE INDEX idx_salt_returns_jid ON salt_returns (jid);
    CREATE INDEX idx_salt_returns_fun ON salt_returns (fun);
    CREATE INDEX idx_salt_returns_updated ON salt_returns (alter_time);

    --
    -- Table structure for table `salt_events`
    --

    DROP TABLE IF EXISTS salt_events;
    DROP SEQUENCE IF EXISTS seq_salt_events_id;
    CREATE SEQUENCE seq_salt_events_id;
    CREATE TABLE salt_events (
        id BIGINT NOT NULL UNIQUE DEFAULT nextval('seq_salt_events_id'),
        tag varchar(255) NOT NULL,
        data text NOT NULL,
        alter_time TIMESTAMP WITH TIME ZONE DEFAULT NOW(),
        master_id varchar(255) NOT NULL
    );

    CREATE INDEX idx_salt_events_tag on salt_events (tag);

    EOF

Required python modules: psycopg2

To use the postgres returner, append '--return postgres' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return postgres

To use the alternative configuration, append '--return_config alternative' to the salt command.

.. versionadded:: 2015.5.0

.. code-block:: bash

    salt '*' test.ping --return postgres --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return postgres --return_kwargs '{"db": "another-salt"}'

"""

import logging
import sys
from contextlib import contextmanager

import salt.exceptions
import salt.returners
import salt.utils.data
import salt.utils.json

try:
    import psycopg2

    HAS_POSTGRES = True
except ImportError:
    HAS_POSTGRES = False

__virtualname__ = "postgres"

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_POSTGRES:
        return False, "Could not import postgres returner; psycopg2 is not installed."
    return __virtualname__


def _get_options(ret=None):
    """
    Get the postgres options from salt.
    """
    defaults = {
        "host": "localhost",
        "user": "salt",
        "passwd": "salt",
        "db": "salt",
        "port": 5432,
    }

    attrs = {
        "host": "host",
        "user": "user",
        "passwd": "passwd",
        "db": "db",
        "port": "port",
    }

    _options = salt.returners.get_returner_options(
        f"returner.{__virtualname__}",
        ret,
        attrs,
        __salt__=__salt__,
        __opts__=__opts__,
        defaults=defaults,
    )
    # Ensure port is an int
    if "port" in _options:
        _options["port"] = int(_options["port"])
    return _options


@contextmanager
def _get_serv(ret=None, commit=False):
    """
    Return a Pg cursor
    """
    _options = _get_options(ret)
    try:
        conn = psycopg2.connect(
            host=_options.get("host"),
            user=_options.get("user"),
            password=_options.get("passwd"),
            database=_options.get("db"),
            port=_options.get("port"),
        )

    except psycopg2.OperationalError as exc:
        raise salt.exceptions.SaltMasterError(
            f"postgres returner could not connect to database: {exc}"
        )

    cursor = conn.cursor()

    try:
        yield cursor
    except psycopg2.DatabaseError as err:
        error = err.args
        sys.stderr.write(str(error))
        cursor.execute("ROLLBACK")
        raise
    else:
        if commit:
            cursor.execute("COMMIT")
        else:
            cursor.execute("ROLLBACK")
    finally:
        conn.close()


def returner(ret):
    """
    Return data to a postgres server
    """
    try:
        with _get_serv(ret, commit=True) as cur:
            sql = """INSERT INTO salt_returns
                    (fun, jid, return, id, success, full_ret)
                    VALUES (%s, %s, %s, %s, %s, %s)"""
            cleaned_return = salt.utils.data.decode(ret)
            cur.execute(
                sql,
                (
                    ret["fun"],
                    ret["jid"],
                    salt.utils.json.dumps(cleaned_return["return"]),
                    ret["id"],
                    ret.get("success", False),
                    salt.utils.json.dumps(cleaned_return),
                ),
            )
    except salt.exceptions.SaltMasterError:
        log.critical(
            "Could not store return with postgres returner. PostgreSQL server"
            " unavailable."
        )


def event_return(events):
    """
    Return event to Pg server

    Requires that configuration be enabled via 'event_return'
    option in master config.
    """
    with _get_serv(events, commit=True) as cur:
        for event in events:
            tag = event.get("tag", "")
            data = event.get("data", "")
            sql = """INSERT INTO salt_events (tag, data, master_id)
                     VALUES (%s, %s, %s)"""
            cur.execute(sql, (tag, salt.utils.json.dumps(data), __opts__["id"]))


def save_load(jid, load, minions=None):  # pylint: disable=unused-argument
    """
    Save the load to the specified jid id
    """
    with _get_serv(commit=True) as cur:

        sql = """INSERT INTO jids
               (jid, load)
                VALUES (%s, %s)"""

        json_data = salt.utils.json.dumps(salt.utils.data.decode(load))
        try:
            cur.execute(sql, (jid, json_data))
        except psycopg2.IntegrityError:
            # https://github.com/saltstack/salt/issues/22171
            # Without this try/except we get tons of duplicate entry errors
            # which result in job returns not being stored properly
            pass


def save_minions(jid, minions, syndic_id=None):  # pylint: disable=unused-argument
    """
    Included for API consistency
    """


def get_load(jid):
    """
    Return the load data that marks a specified jid
    """
    with _get_serv(ret=None, commit=True) as cur:
        sql = """SELECT load FROM jids WHERE jid = %s;"""
        cur.execute(sql, (jid,))
        data = cur.fetchone()
        if data:
            return salt.utils.json.loads(data[0])
        return {}


def get_jid(jid):
    """
    Return the information returned when the specified job id was executed
    """
    with _get_serv(ret=None, commit=True) as cur:

        sql = """SELECT id, full_ret FROM salt_returns
                WHERE jid = %s"""

        cur.execute(sql, (jid,))
        data = cur.fetchall()
        ret = {}
        if data:
            for minion, full_ret in data:
                ret[minion] = salt.utils.json.loads(full_ret)
        return ret


def get_fun(fun):
    """
    Return a dict of the last function called for all minions
    """
    with _get_serv(ret=None, commit=True) as cur:

        sql = """SELECT s.id,s.jid, s.full_ret
                FROM salt_returns s
                JOIN ( SELECT MAX(`jid`) as jid
                    from salt_returns GROUP BY fun, id) max
                ON s.jid = max.jid
                WHERE s.fun = %s
                """

        cur.execute(sql, (fun,))
        data = cur.fetchall()

        ret = {}
        if data:
            for minion, _, full_ret in data:
                ret[minion] = salt.utils.json.loads(full_ret)
        return ret


def get_jids():
    """
    Return a list of all job ids
    """
    with _get_serv(ret=None, commit=True) as cur:

        sql = """SELECT jid, load
                FROM jids"""

        cur.execute(sql)
        data = cur.fetchall()
        ret = {}
        for jid, load in data:
            ret[jid] = salt.utils.jid.format_jid_instance(
                jid, salt.utils.json.loads(load)
            )
        return ret


def get_minions():
    """
    Return a list of minions
    """
    with _get_serv(ret=None, commit=True) as cur:

        sql = """SELECT DISTINCT id
                FROM salt_returns"""

        cur.execute(sql)
        data = cur.fetchall()
        ret = []
        for minion in data:
            ret.append(minion[0])
        return ret


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

OPENSHIFT_CLUSTERQUOTA_NAME: Final = "openshift.clusterquota.name"
"""
The name of the cluster quota.
"""

OPENSHIFT_CLUSTERQUOTA_UID: Final = "openshift.clusterquota.uid"
"""
The UID of the cluster quota.
"""



# This module is part of GitPython and is released under the
# 3-Clause BSD License: https://opensource.org/license/bsd-3-clause/

"""Index utilities."""

__all__ = ["TemporaryFileSwap", "post_clear_cache", "default_index", "git_working_dir"]

import contextlib
from functools import wraps
import os
import os.path as osp
import struct
import tempfile
from types import TracebackType

# typing ----------------------------------------------------------------------

from typing import Any, Callable, TYPE_CHECKING, Optional, Type, cast

from git.types import Literal, PathLike, _T

if TYPE_CHECKING:
    from git.index import IndexFile

# ---------------------------------------------------------------------------------

# { Aliases
pack = struct.pack
unpack = struct.unpack
# } END aliases


class TemporaryFileSwap:
    """Utility class moving a file to a temporary location within the same directory and
    moving it back on to where on object deletion."""

    __slots__ = ("file_path", "tmp_file_path")

    def __init__(self, file_path: PathLike) -> None:
        self.file_path = file_path
        dirname, basename = osp.split(file_path)
        fd, self.tmp_file_path = tempfile.mkstemp(prefix=basename, dir=dirname)
        os.close(fd)
        with contextlib.suppress(OSError):  # It may be that the source does not exist.
            os.replace(self.file_path, self.tmp_file_path)

    def __enter__(self) -> "TemporaryFileSwap":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Literal[False]:
        if osp.isfile(self.tmp_file_path):
            os.replace(self.tmp_file_path, self.file_path)
        return False


# { Decorators


def post_clear_cache(func: Callable[..., _T]) -> Callable[..., _T]:
    """Decorator for functions that alter the index using the git command.

    When a git command alters the index, this invalidates our possibly existing entries
    dictionary, which is why it must be deleted to allow it to be lazily reread later.
    """

    @wraps(func)
    def post_clear_cache_if_not_raised(self: "IndexFile", *args: Any, **kwargs: Any) -> _T:
        rval = func(self, *args, **kwargs)
        self._delete_entries_cache()
        return rval

    # END wrapper method

    return post_clear_cache_if_not_raised


def default_index(func: Callable[..., _T]) -> Callable[..., _T]:
    """Decorator ensuring the wrapped method may only run if we are the default
    repository index.

    This is as we rely on git commands that operate on that index only.
    """

    @wraps(func)
    def check_default_index(self: "IndexFile", *args: Any, **kwargs: Any) -> _T:
        if self._file_path != self._index_path():
            raise AssertionError(
                "Cannot call %r on indices that do not represent the default git index" % func.__name__
            )
        return func(self, *args, **kwargs)

    # END wrapper method

    return check_default_index


def git_working_dir(func: Callable[..., _T]) -> Callable[..., _T]:
    """Decorator which changes the current working dir to the one of the git
    repository in order to ensure relative paths are handled correctly."""

    @wraps(func)
    def set_git_working_dir(self: "IndexFile", *args: Any, **kwargs: Any) -> _T:
        cur_wd = os.getcwd()
        os.chdir(cast(PathLike, self.repo.working_tree_dir))
        try:
            return func(self, *args, **kwargs)
        finally:
            os.chdir(cur_wd)
        # END handle working dir

    # END wrapper

    return set_git_working_dir


# } END decorators
